"""Advanced fuel cell module."""

# clean
import os
from dataclasses import dataclass
import math

from typing import Any, List
import copy
from dataclasses_json import dataclass_json

import pandas as pd
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput, ConfigBase, DisplayConfig
from hisim import loadtypes as lt

from hisim.components.configuration import PhysicsConfig
from hisim import utils
from hisim.simulationparameters import SimulationParameters
from hisim import log


__authors__ = "Frank Burkrad, Maximilian Hillen,"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class CHPConfig(ConfigBase):
    """CHP Config class."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return CHP.get_full_classname()

    building_name: str
    name: str
    min_operation_time: float
    min_idle_time: float
    gas_type: str
    operating_mode: str
    is_modulating: bool
    p_el_min: float  # [W]
    p_th_min: float  # [W]
    eff_el_min: float  # [-]
    eff_th_min: float  # [-]
    mass_flow_max: float  # kg/s
    p_el_max: float  # [W]
    p_th_max: float  # [W]
    eff_el_max: float  # [-]
    eff_th_max: float  # [-]
    temperature_max: float
    delta_temperature: float

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default config."""
        config = CHPConfig(
            building_name=building_name,
            name="CHP",
            min_operation_time=60,
            min_idle_time=15,
            gas_type="Hydrogen",
            operating_mode="both",
            is_modulating=True,
            p_el_min=2_000,
            p_th_min=3_000,
            eff_el_min=0.2,
            eff_th_min=0.5,
            mass_flow_max=0.011,
            p_el_max=3_000,
            p_th_max=4_000,
            eff_el_max=0.4,
            eff_th_max=0.55,
            temperature_max=80,
            delta_temperature=10,
        )
        return config


class CHPConfigAdvanced:
    """CHP config advanced class."""

    def __init__(self) -> None:
        """Initialize the class."""
        # Remark: moved the whole class body into the __init__ function to avoid errors if the file read below
        # does not exist.

        # system_name = "BlueGEN15"
        # system_name = "Dachs 0.8"
        # system_name = "Test_KWK"
        # system_name = "Dachs G2.9"
        # system_name = "HOMER"
        system_name = "BlueGen BG15"

        dataframe = pd.read_excel(
            os.path.join(utils.HISIMPATH["chp_system"], "mock_up_efficiencies.xlsx"),
            index_col=0,
        )

        df_specific = dataframe.loc[str(system_name)]

        if str(df_specific["is_modulating"]) == "Yes":
            self.is_modulating = True
            self.p_el_min = df_specific["P_el_min"]
            self.p_th_min = df_specific["P_th_min"]
            self.p_total_min = df_specific["P_total_min"]
            self.eff_el_min = df_specific["eff_el_min"]
            self.eff_th_min = df_specific["eff_th_min"]

        elif str(df_specific["is_modulating"]) == "No":
            self.is_modulating = False
        else:
            log.error("Modulation is not defined. Modulation must be 'Yes' or 'No'")
            raise ValueError

        self.p_el_max = df_specific["P_el_max"]
        self.p_th_max = df_specific["P_th_max"]
        self.p_total_max = df_specific["P_total_max"]  # maximum fuel consumption
        self.eff_el_max = df_specific["eff_el_max"]
        self.eff_th_max = df_specific["eff_th_max"]
        self.mass_flow_max = df_specific["mass_flow (dT=20°C)"]
        self.temperature_max = df_specific["temperature_max"]
        self.delta_temperature = 10


class CHPState:
    """CHP state class."""

    def __init__(self, start_timestep=None, electricity_output=0.0, cycle_number=None):
        """Initialize the class."""
        self.start_timestep = start_timestep
        self.electricity_output = electricity_output
        self.cycle_number = cycle_number
        if self.electricity_output == 0.0:
            self.activation = 0
        elif self.electricity_output > 0.0:
            self.activation = 1
        else:
            raise Exception("Impossible CHPState.")


class CHP(Component):
    """CHP class.

    Simulate chp efficiency (cop) as well as electrical (p_el) &
    thermal power (p_th), massflow (m_dot) and output temperature (t_out).
    """

    # Inputs
    ControlSignal = "ControlSignal"  # at which Procentage is the CHP modulating [0..1]
    MassflowInputTemperature = "MassflowInputTemperature"
    ElectricityFromCHPTarget = "ElectricityFromCHPTarget"
    HydrogenNotReleased = "HydrogenNotReleased"
    # OperatingModelSignal="OperatingModelSignal" #-->Wärme oder Stromgeführt. Nötig?

    # Output
    MassflowOutput = "Hot Water Energy Output"
    MassflowOutputTemperature = "MassflowOutputTemperature"
    ElectricityOutput = "ElectricityOutput"
    GasDemandTarget = "GasDemandTarget"
    NumberofCycles = "NumberofCycles"
    ThermalOutputPower = "ThermalOutputPower"
    GasDemandReal = "GasDemandReal"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: CHPConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.chp_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.min_operation_time = self.chp_config.min_operation_time
        self.min_idle_time = self.chp_config.min_idle_time
        self.gas_type = self.chp_config.gas_type  # Gas Type can be "Hydrogen" or "Methan"
        self.operating_mode = self.chp_config.operating_mode  # operating_mode=["both","heat","electricity"]

        self.number_of_cycles = 0
        self.number_of_cycles_previous = copy.deepcopy(self.number_of_cycles)
        self.state = CHPState(start_timestep=int(0), cycle_number=0)
        self.previous_state = copy.deepcopy(self.state)

        # the 3600 comes from Normalised chp from p_el_max=3600. Look up chp_system_lib for more information
        self.p_el_max = self.chp_config.p_el_max
        usually_p_el_max = self.chp_config.p_el_max
        self.p_th_max = self.chp_config.p_th_max * (self.p_el_max / usually_p_el_max)
        self.p_th_min = self.chp_config.p_th_min
        self.p_el_min = self.chp_config.p_el_min

        if self.p_el_max < self.p_el_min or self.p_th_max < self.p_th_min:
            self.p_el_max = self.p_el_min + 100
            self.p_th_max = self.p_th_max + 100

        self.mass_flow_max = self.chp_config.mass_flow_max * (self.p_el_max / usually_p_el_max)
        if self.mass_flow_max < self.chp_config.mass_flow_max:
            self.mass_flow_max = self.chp_config.mass_flow_max
        self.eff_th_min: float = self.chp_config.eff_th_min
        self.eff_th_max = self.chp_config.eff_th_max
        self.eff_el_min: float = self.chp_config.eff_el_min
        self.eff_el_max = self.chp_config.eff_el_max
        self.temperature_max = self.chp_config.temperature_max

        self.delta_t = self.chp_config.delta_temperature

        # Inputs
        self.control_signal_channel: ComponentInput = self.add_input(
            self.component_name,
            CHP.ControlSignal,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            False,
        )
        # self.operating_mode_signal: ComponentInput = self.add_input(self.ComponentName, CHP.OperatingModelSignal, lt.LoadTypes.Gas, lt.Units.Percent, True)
        self.mass_inp_temp_channel: ComponentInput = self.add_input(
            self.component_name,
            CHP.MassflowInputTemperature,
            lt.LoadTypes.WATER,
            lt.Units.CELSIUS,
            False,
        )
        self.electricity_target_channel: ComponentInput = self.add_input(
            self.component_name,
            CHP.ElectricityFromCHPTarget,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            False,
        )
        self.hydrogen_not_released_channel: ComponentInput = self.add_input(
            self.component_name,
            CHP.HydrogenNotReleased,
            lt.LoadTypes.GAS,
            lt.Units.KG,
            False,
        )

        # Outputs
        self.mass_out_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.MassflowOutput,
            lt.LoadTypes.WATER,
            lt.Units.KG_PER_SEC,
            output_description=f"here a description for {self.MassflowOutput} will follow.",
        )
        self.mass_out_temp_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.MassflowOutputTemperature,
            lt.LoadTypes.WATER,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.MassflowOutputTemperature} will follow.",
        )
        self.gas_demand_target_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.GasDemandTarget,
            lt.LoadTypes.GAS,
            lt.Units.KG_PER_SEC,
            output_description=f"here a description for {self.GasDemandTarget} will follow.",
        )
        self.el_power_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.ElectricityOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description=f"here a description for CHP {self.ElectricityOutput} will follow.",
        )
        self.number_of_cycles_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.NumberofCycles,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description=f"here a description for CHP {self.NumberofCycles} will follow.",
        )
        self.th_power_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.ThermalOutputPower,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for CHP {self.ThermalOutputPower} will follow.",
        )
        self.gas_demand_real_used_channel: ComponentOutput = self.add_output(
            self.component_name,
            CHP.GasDemandReal,
            lt.LoadTypes.GAS,
            lt.Units.KG_PER_SEC,
            output_description=f"here a description for {self.GasDemandReal} will follow.",
        )

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = copy.deepcopy(self.state)
        self.number_of_cycles_previous = self.number_of_cycles

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = copy.deepcopy(self.previous_state)
        self.number_of_cycles = self.number_of_cycles_previous

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doubelchecks."""
        pass

    def simulate_chp(self, control_signal: float, stsv: SingleTimeStepValues, timestep: int) -> Any:
        """Simualtes the component."""

        specific_heat_capacity_water = 4182
        # Calculation.Electric Energy deliverd
        # CHP is on
        if self.state.activation != 0:
            self.number_of_cycles = self.state.cycle_number
            # Checks if the minimum running time has been reached

            # Minimium running time has been reached and the CHP wants to shut off -->so it shuts off
            if timestep >= self.state.start_timestep + self.min_operation_time and control_signal == 0.0:
                # all Outputs zero
                mass_out_temp: float = 0
                mass_out: float = 0
                el_power: float = 0
                th_power: float = 0
                eff_el_real: float = 0
                eff_th_real: float = 0

                self.state = CHPState(
                    start_timestep=timestep,
                    cycle_number=self.number_of_cycles,
                    electricity_output=el_power,
                )

                stsv.set_output_value(self.mass_out_temp_channel, mass_out_temp)  # mass output temp
                stsv.set_output_value(self.mass_out_channel, mass_out)  # mass output flow
                stsv.set_output_value(self.el_power_channel, el_power)  # mass output flow
                stsv.set_output_value(self.th_power_channel, th_power)  # mass output flow

                # zu ändern, da gas_demand=!gas_power
                stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)
                return el_power, th_power, eff_el_real, eff_th_real

            # Minimium running time has not been reached and the CHP wants to shut off -->so it won't shut off
            if timestep < self.state.start_timestep + self.min_operation_time and control_signal == 0:
                # CHP doesn't want to run but has to, therefore is going to run in minimum power

                maximum_power_th: float = self.p_th_min
                eff_th_real = self.eff_th_min
                eff_el_real = self.eff_el_min
                maximum_power_el: float = self.p_el_min

                th_power = maximum_power_th * eff_th_real
                el_power = maximum_power_el * eff_el_real

                mass_out_temp = self.delta_t + stsv.get_input_value(self.mass_inp_temp_channel)
                mass_out = th_power / (specific_heat_capacity_water * self.delta_t)

                if mass_out > self.mass_flow_max:
                    mass_out = self.mass_flow_max
                    mass_out_temp = stsv.get_input_value(self.mass_inp_temp_channel) + th_power / (
                        mass_out * specific_heat_capacity_water
                    )

            # CHP doens't want to shut off and its activated--> so its stays activated
            else:
                # Calculate Eff_th
                d_eff_th = self.eff_th_max - self.eff_th_min

                if control_signal * self.p_th_max < self.p_th_min:
                    maximum_power_th = self.p_th_min
                    eff_th_real = self.eff_th_min
                else:
                    maximum_power_th = control_signal * self.p_th_max
                    eff_th_real = self.eff_th_min + d_eff_th * control_signal

                # Calculate Eff_el
                d_eff_el = self.eff_el_max - self.eff_el_min

                if control_signal * self.p_el_max < self.p_el_min:
                    maximum_power_el = self.p_el_min
                    eff_el_real = self.eff_el_min
                else:
                    maximum_power_el = control_signal * self.p_el_max
                    eff_el_real = self.eff_el_min + d_eff_el * control_signal

                th_power = maximum_power_th * eff_th_real
                el_power = maximum_power_el * eff_el_real

                mass_out_temp = self.delta_t + stsv.get_input_value(self.mass_inp_temp_channel)
                mass_out = th_power / (specific_heat_capacity_water * self.delta_t)

                if mass_out > self.mass_flow_max:
                    mass_out = self.mass_flow_max
                    mass_out_temp = stsv.get_input_value(self.mass_inp_temp_channel) + th_power / (
                        mass_out * specific_heat_capacity_water
                    )

            th_power = (
                (mass_out_temp - stsv.get_input_value(self.mass_inp_temp_channel))
                * specific_heat_capacity_water
                * mass_out
            )
            stsv.set_output_value(self.th_power_channel, th_power)  # ThermalPowerOutput
            stsv.set_output_value(self.mass_out_temp_channel, mass_out_temp)  # mass output temp
            stsv.set_output_value(self.mass_out_channel, mass_out)  # mass output flow
            stsv.set_output_value(self.el_power_channel, el_power)  # mass output flow
            # zu ändern, da gas_demand=!gas_power
            stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)
            return el_power, th_power, eff_el_real, eff_th_real
            # run in power of control_signal

        # CHP is Off
        # CHP wants to start and waited long enough since last start--> so it starts
        if control_signal != 0 and (timestep >= self.state.start_timestep + self.min_idle_time):
            self.number_of_cycles = self.number_of_cycles + 1
            number_of_cycles = self.number_of_cycles

            # Calculate Eff_th
            d_eff_th = self.eff_th_max - self.eff_th_min

            if control_signal * self.p_th_max < self.p_th_min:
                maximum_power_th = self.p_th_min
                eff_th_real = self.eff_th_min
            else:
                maximum_power_th = control_signal * self.p_th_max
                eff_th_real = self.eff_th_min + d_eff_th * control_signal

            # Calculate Eff_el
            d_eff_el = self.eff_el_max - self.eff_el_min

            if control_signal * self.p_el_max < self.p_el_min:
                maximum_power_el = self.p_el_min
                eff_el_real = self.eff_el_min
            else:
                maximum_power_el = control_signal * self.p_el_max
                eff_el_real = self.eff_el_min + d_eff_el * control_signal

            th_power = maximum_power_th * eff_th_real
            el_power = maximum_power_el * eff_el_real

            mass_out_temp = self.delta_t + stsv.get_input_value(self.mass_inp_temp_channel)
            mass_out = th_power / (specific_heat_capacity_water * self.delta_t)

            if mass_out > self.mass_flow_max:
                mass_out = self.mass_flow_max
                mass_out_temp = stsv.get_input_value(self.mass_inp_temp_channel) + th_power / (
                    mass_out * specific_heat_capacity_water
                )

            self.state = CHPState(
                start_timestep=timestep,
                electricity_output=el_power,
                cycle_number=number_of_cycles,
            )

        # CHP wants to starts but didn't wait long enough since last start -> so it won't start
        else:
            # all Outputs should be zero, because CHP can't start
            mass_out_temp = 0
            mass_out = 0
            el_power = 0
            th_power = 0
            eff_el_real = 0
            eff_th_real = 0

        stsv.set_output_value(self.th_power_channel, th_power)  # ThermalPowerOutput
        stsv.set_output_value(self.mass_out_temp_channel, mass_out_temp)  # mass output temp
        stsv.set_output_value(self.mass_out_channel, mass_out)  # mass output flow
        stsv.set_output_value(self.el_power_channel, el_power)  # mass output flow
        # zu ändern, da gas_demand=!gas_power
        stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)

        return el_power, th_power, eff_el_real, eff_th_real

    def calculate_control_signal(self, stsv: SingleTimeStepValues) -> float:  # pylint: disable=R0911
        """Calculate control signal."""
        if (stsv.get_input_value(self.electricity_target_channel)) < 30:
            control_signal: float = 0
            return control_signal
        if (stsv.get_input_value(self.electricity_target_channel)) < self.p_el_min * self.eff_el_min:
            control_signal = 0.4
            return control_signal
        if (stsv.get_input_value(self.electricity_target_channel)) > self.p_el_max * self.eff_el_max:
            control_signal = 1
            return control_signal

        x_1 = (
            -self.p_el_max
            - math.sqrt(
                (self.p_el_max * self.eff_el_min) ** 2
                + 4
                * (
                    stsv.get_input_value(self.electricity_target_channel)
                    * self.p_el_max
                    * (self.eff_el_max - self.eff_el_min)
                )
            )
        ) / (2 * self.p_el_max * (self.eff_el_max - self.eff_el_min))
        x_2 = (
            -self.p_el_max
            + math.sqrt(
                (self.p_el_max * self.eff_el_min) ** 2
                + 4
                * (
                    stsv.get_input_value(self.electricity_target_channel)
                    * self.p_el_max
                    * (self.eff_el_max - self.eff_el_min)
                )
            )
        ) / (2 * self.p_el_max * (self.eff_el_max - self.eff_el_min))
        if 0 < x_1 < 1:
            if 0 < x_2 < 1:
                if x_1 < x_2:
                    return x_2

                return x_1

            return x_1

        return x_2

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        control_signal: float = -1
        if self.operating_mode == "heat":
            control_signal = stsv.get_input_value(self.control_signal_channel)
        elif self.operating_mode == "electricity":
            control_signal = self.calculate_control_signal(stsv)
            if control_signal > 1 or control_signal < 0:
                control_signal = 1
        elif self.operating_mode == "both":
            control_signal = self.calculate_control_signal(stsv)
            if control_signal <= stsv.get_input_value(self.control_signal_channel):
                control_signal = stsv.get_input_value(self.control_signal_channel)
            else:
                if control_signal > 1 or control_signal < 0:
                    control_signal = 1

        if control_signal > 1:
            raise Exception("Expected a control signal between 0 and 1")
        if control_signal < 0:
            raise Exception("Expected a control signal between 0 and 1")

        el_power, th_power, eff_el_real, eff_th_real = self.simulate_chp(
            control_signal=control_signal, stsv=stsv, timestep=timestep
        )
        # Check if enough hydrogen is in the tank
        gas_demand_real_used = 0
        if el_power == 0 and th_power == 0:
            gas_demand_target = 0
            gas_demand_real_used = 0
        else:
            if self.gas_type == "Hydrogen":
                gas_demand_target = ((el_power / eff_el_real) + (th_power / eff_th_real)) / (
                    PhysicsConfig.hydrogen_specific_fuel_value_per_kg
                )
                if stsv.get_input_value(self.hydrogen_not_released_channel) == 0:
                    # Gas Demand can completly be charged from storage
                    gas_demand_real_used = gas_demand_target

                elif stsv.get_input_value(self.hydrogen_not_released_channel) < 0:
                    log.warning("Fault, bc. GasDemandpossible>gasdemandtarget @ " + str(timestep))

                elif stsv.get_input_value(self.hydrogen_not_released_channel) > 0:
                    # not enough Gas for running CHP on power demanded/calculated before
                    # to simplify, turn of CHP complelty, also when minimum running time isn't reached, bec. no Hydrogen is there

                    stsv.set_output_value(self.th_power_channel, 0)  # ThermalPowerOutput
                    stsv.set_output_value(self.mass_out_temp_channel, 0)  # mass output temp
                    stsv.set_output_value(self.mass_out_channel, 0)  # mass output flow
                    stsv.set_output_value(self.el_power_channel, 0)  # mass output flow
                    self.number_of_cycles = self.state.cycle_number
                    el_power = 0
                    self.state = CHPState(
                        start_timestep=timestep,
                        cycle_number=self.number_of_cycles,
                        electricity_output=el_power,
                    )

                    stsv.set_output_value(self.number_of_cycles_channel, self.number_of_cycles)
                    gas_demand_real_used = 0
            elif self.gas_type == "Methan":
                gas_demand_target = (
                    (el_power / eff_el_real) + (th_power / eff_th_real)
                ) / PhysicsConfig.natural_gas_specific_fuel_value_per_kg
                gas_demand_real_used = gas_demand_target
            else:
                raise Exception("No Gas chosen which is integrated in System")

        stsv.set_output_value(self.gas_demand_target_channel, gas_demand_target)  # CHP runs with
        stsv.set_output_value(self.gas_demand_real_used_channel, gas_demand_real_used)  # ThermalPowerOutput

    def write_to_report(self) -> List[str]:
        """Write to report."""
        lines = []
        for config_string in self.chp_config.get_string_dict():
            lines.append(config_string)
        lines.append("Component Name: " + str(self.component_name))
        lines.append("Name: CHP")
        lines.append("Min Operation Time [Sec]: " + str(self.min_operation_time))
        lines.append("Min Idle Time [Sec]: " + str(self.min_idle_time))
        lines.append("Gas Type: " + str(self.gas_type))
        lines.append("Operating Mode: " + str(self.operating_mode))
        lines.append("P_el_max [P]: " + str(self.p_el_max))
        lines.append("P_el_min [P]: " + str(self.p_el_min))
        lines.append("Eff_el_min: " + str(self.eff_el_min))
        lines.append("Eff_el_max: " + str(self.eff_el_max))
        lines.append("Mass Flow Max: " + str(self.mass_flow_max))
        lines.append("P_th_min [P]: " + str(self.p_th_min))
        lines.append("P_th_max [P]: " + str(self.p_th_max))
        lines.append("Eff_th_min: " + str(self.eff_th_min))
        lines.append("Eff_th_max: " + str(self.eff_th_max))
        lines.append("Max Temperature [°C]: " + str(self.temperature_max))
        lines.append("Delta T [°C]: " + str(self.delta_t))

        return lines



"""Result Data Processing and Plotting for Scenario Comparison."""


import os
from enum import Enum
from dataclasses import dataclass
from typing import Dict, Tuple, Optional, List
import pandas as pd
from ordered_set import OrderedSet
from hisim import log


class ScenarioDataProcessing:

    """ScenarioDataProcessing class."""

    @staticmethod
    def get_dataframe_and_create_pandas_dataframe_for_all_data(
        filepath_of_aggregated_dataframe: str,
        dict_of_scenarios_to_check: Optional[Dict[str, List[str]]],
        variables_to_check: List[str],
        data_format_type: str,
    ) -> Tuple[pd.DataFrame, List[str]]:
        """Get csv data and create dataframes with the filtered and procesed scenario data."""

        # if time_resolution_of_data_set == ResultDataTypeEnum.HOURLY.name:
        #     kind_of_data_set = "hourly"
        # elif time_resolution_of_data_set == ResultDataTypeEnum.YEARLY.name:
        #     kind_of_data_set = "yearly"
        # elif time_resolution_of_data_set == ResultDataTypeEnum.DAILY.name:
        #     kind_of_data_set = "daily"
        # elif time_resolution_of_data_set == ResultDataTypeEnum.MONTHLY.name:
        #     kind_of_data_set = "monthly"
        # else:
        #     raise ValueError("This kind of data was not found in the datacollectorenum class.")
        if not os.path.isfile(filepath_of_aggregated_dataframe):
            raise FileExistsError(f"The file {filepath_of_aggregated_dataframe} could not be found.")
        log.information(f"Read aggregated dataframe with all HiSim results from {filepath_of_aggregated_dataframe}.")

        # data_file_path = os.path.join(data_folder_path, f"*{kind_of_data_set}*.{data_format_type.lower()}")
        # list_of_possible_data_files = glob.glob(data_file_path)
        # file: str = ""
        # if not list_of_possible_data_files:
        #     raise FileExistsError(f"No file could be found in this path {data_file_path}.")
        # if len(list_of_possible_data_files) > 1:
        #     raise ValueError(f"The file path {data_file_path} should not contain more than one file.")

        # file = list_of_possible_data_files[0]
        if data_format_type == DataFormatEnum.CSV.name:
            # file_df = pd.read_csv(filepath_or_buffer=f"{file}.gz", header=[0, 1], compression="gzip")
            file_df = pd.read_csv(filepath_or_buffer=filepath_of_aggregated_dataframe, header=[0, 1])
        elif data_format_type == DataFormatEnum.XLSX.name:
            file_df = pd.read_excel(filepath_of_aggregated_dataframe, header=[0, 1], index_col=0, sheet_name="Sheet1")
        else:
            raise ValueError(f"Only data format types xlsx or csv are implemented. Here it is {data_format_type}.")

        if dict_of_scenarios_to_check is not None and dict_of_scenarios_to_check != {}:

            file_df = ScenarioDataProcessing.check_if_scenario_exists_and_filter_dataframe_for_scenarios_dict(
                data_frame=file_df, dict_of_scenarios_to_check=dict_of_scenarios_to_check,
            )

        return (
            file_df,
            variables_to_check,
        )

    @staticmethod
    def filter_pandas_dataframe_according_to_output_variable(
        dataframe: pd.DataFrame, variable_to_check: str
    ) -> pd.DataFrame:
        """Filter pandas dataframe according to variable."""

        filtered_dataframe = dataframe.loc[dataframe[("Output", "variable")] == variable_to_check]
        if filtered_dataframe.empty:
            print("The dataframe contains the following variables: ", set(list(dataframe[("Output", "variable")])))
            # raise ValueError(
            print(
                f"The filtered dataframe is empty. The dataframe did not contain the variable {variable_to_check}. Check the list above."
            )
        return filtered_dataframe

    @staticmethod
    def get_statistics_of_data_and_write_to_excel(
        filtered_data: pd.DataFrame, x_and_y_plot_data: pd.DataFrame, path_to_save: str, kind_of_data_set: str,
    ) -> None:
        """Use pandas describe method to get statistical values of certain data."""

        # create a excel writer object
        with pd.ExcelWriter(  # pylint: disable=abstract-class-instantiated
            path=os.path.join(path_to_save, f"{kind_of_data_set}_stats.xlsx"), mode="w",
        ) as writer:
            # statistical data
            statistical_data = filtered_data.describe()
            statistical_data.to_excel(excel_writer=writer, sheet_name="statistics")
            # write also x and y data which is plotted
            x_and_y_plot_data.to_excel(excel_writer=writer, sheet_name="plotted mean data")

    @staticmethod
    def aggregate_all_values_for_one_scenario_and_rename(
        dataframe: pd.DataFrame, list_of_scenarios_to_check: List, column_name_to_check: str,
    ) -> pd.DataFrame:
        """Check for one scenario."""

        # Create a filtered DataFrame for each scenario and store in a list
        filtered_dfs = []
        # Iterate through each scenario in the list
        for scenario_to_check in list_of_scenarios_to_check:
            log.information("Filtering dataframe for scenario: " + scenario_to_check)

            if isinstance(scenario_to_check, str):
                # Filter DataFrame rows where the specified column contains the scenario string
                filtered_df = dataframe[
                    dataframe[("Input", column_name_to_check)].str.contains(scenario_to_check, na=False)
                ]
            else:
                # Filter DataFrame rows where the specified column matches the scenario value (for numeric scenarios)
                filtered_df = dataframe[dataframe[("Input", column_name_to_check)] == scenario_to_check]

            # Rename scenarios according to scenario_to_check
            filtered_df[("Input", "scenario")] = scenario_to_check

            # Add the filtered DataFrame to the list
            filtered_dfs.append(filtered_df)

        # Concatenate all filtered DataFrames
        concat_df = pd.concat(filtered_dfs, ignore_index=True)

        return concat_df

    @staticmethod
    def check_if_scenario_exists_and_filter_dataframe_for_scenarios_dict(
        data_frame: pd.DataFrame, dict_of_scenarios_to_check: Dict[str, List[str]],
    ) -> pd.DataFrame:
        """Check if scenario exists and filter dataframe for scenario."""

        concat_df = data_frame
        filter_level_index = 0
        # Iterate over the dictionary of scenarios
        for scenario_to_check_key, list_of_scenarios_to_check in dict_of_scenarios_to_check.items():
            log.information(
                f"Scenarios to check are {list_of_scenarios_to_check} in Input column {scenario_to_check_key}."
            )
            concat_df = ScenarioDataProcessing.aggregate_all_values_for_one_scenario_and_rename(
                dataframe=concat_df,
                list_of_scenarios_to_check=list_of_scenarios_to_check,
                column_name_to_check=scenario_to_check_key,
            )
            filter_level_index += 1

        if filter_level_index > 1:
            raise ValueError(
                f"Filter level index is {filter_level_index} but should be 1 at max."
                " This means your dict_of_scenarios_to_check should only have one scenario_key to filter."
            )

        return concat_df

    @staticmethod
    def take_mean_values_of_scenarios(
        filtered_data: pd.DataFrame, time_resolution_of_data_set: str
    ) -> Tuple[pd.DataFrame, List[str]]:
        """Get mean values of scenarios."""

        # Dictionary to store the x and y data for plotting
        dict_with_x_and_y_data: Dict[str, List] = {}

        # Get output keys where output values are stored, excluding 'variable' and 'unit'
        output_keys = [key for key in filtered_data["Output"].columns if key not in ("variable", "unit")]

        # The x-axis data is the time or category we are averaging over
        x_data = output_keys

        # Determine if we should take mean values over time
        if len(x_data) == 0:
            raise ValueError(
                f"x data length is 0. This means no keys for output values could be found in the columns of filtered data: {filtered_data.columns}"
            )

        if time_resolution_of_data_set == ResultDataTypeEnum.YEARLY.name and len(x_data) == 1:
            take_mean_time_values_bool = False
        elif time_resolution_of_data_set != ResultDataTypeEnum.YEARLY.name and len(x_data) > 1:
            take_mean_time_values_bool = True
        else:
            raise ValueError("Invalid time resolution and x_data length combination. Check your data and parameters.")

        # Add x_data as a column to the dictionary
        dict_with_x_and_y_data["time"] = x_data

        # Iterate over each unique scenario in the filtered data
        for scenario in OrderedSet(filtered_data[("Input", "scenario")]):
            filtered_data_per_scenario = filtered_data[filtered_data[("Input", "scenario")] == scenario]

            if take_mean_time_values_bool:
                # Compute mean values for each time step
                mean_values = filtered_data_per_scenario["Output"].loc[:, x_data].mean().values
                dict_with_x_and_y_data[f"{scenario}"] = mean_values
            else:
                # Compute the mean value for yearly data
                mean_value = filtered_data_per_scenario[("Output", x_data[0])].mean()
                dict_with_x_and_y_data[f"{scenario}"] = [mean_value]

        # Convert the dictionary to a DataFrame for plotting
        x_and_y_plot_data = pd.DataFrame(dict_with_x_and_y_data)

        return x_and_y_plot_data, output_keys


class ResultDataTypeEnum(Enum):

    """ResultDataTypeEnum class.

    Here it is defined what kind of data you want to collect.
    """

    HOURLY = "hourly"  # hourly not working yet
    DAILY = "daily"
    MONTHLY = "monthly"
    YEARLY = "yearly"


class ResultDataProcessingModeEnum(Enum):

    """ResultDataProcessingModeEnum class.

    Here it is defined what kind of data processing you want to make.
    """

    PROCESS_ALL_DATA = 1
    PROCESS_FOR_DIFFERENT_BUILDING_CODES = 2


class DataFormatEnum(Enum):
    """Class for choosign data format."""

    XLSX = 1
    CSV = 2


@dataclass
class DataInfo:

    """Data info class is an object for storing data in lists of strings."""

    descriptions: List[str]


class OutputVariableEnumClass(Enum):

    """Output variable enum class is for determining variables which will be checked and plotted."""

    KPI_DATA = DataInfo(
        descriptions=[
            "Total electricity consumption",
            "PV production",
            # "Ratio between PV production and total consumption",
            # "Self-consumption",
            # "Self-consumption rate",
            # "Autarky rate",
            "Total energy from grid",
            "Total energy to grid",
            # "Relative electricity demand from grid",
            # "Self-consumption rate according to solar htw berlin",
            "Autarky rate according to solar htw berlin",
            "Investment costs for equipment per simulated period",
            "CO2 footprint for equipment per simulated period",
            "System operational costs for simulated period",
            "System operational emissions for simulated period",
            "Total costs for simulated period",
            "Total CO2 emissions for simulated period",
            # "Temperature deviation of building indoor air temperature being below set temperature 20.0 Celsius",
            # "Minimum building indoor air temperature reached",
            # "Temperature deviation of building indoor air temperature being above set temperature 25.0 Celsius",
            # "Maximum building indoor air temperature reached",
            # "Building heating load",
            "Conditioned floor area",
            # "Rooftop area",
            # "Specific heating load",
            "Specific heating demand according to TABULA",
            # "Thermal output energy of heat distribution system",
            # "Number of SH heat pump cycles",
            # "Seasonal performance factor of SH heat pump",
            # "Seasonal energy efficiency ratio of SH heat pump",
            # "Heating hours of SH heat pump",
            # "Cooling hours of SH heat pump",
            # "Max flow temperature of SH heat pump",
            # "Max return temperature of SH heat pump",
            # "Max temperature difference of SH heat pump",
            # "Min flow temperature of SH heat pump",
            # "Min return temperature of SH heat pump",
            # "Min temperature difference of SH heat pump",
            "Mean flow temperature of SH heat pump",
            "Mean return temperature of SH heat pump",
            # "Heating output energy of SH heat pump",
            # "Cooling output energy of SH heat pump",
            # "Specific heating energy of SH heat pump",
            # "Electrical input energy for heating of SH heat pump",
            # "Electrical input energy for cooling of SH heat pump",
            "Total electrical input energy of SH heat pump",
            "Space heating heat pump electricity from grid",
            # "Relative electricity demand of SH heat pump",
            "DHW heat pump total electricity consumption",
            "Domestic hot water heat pump electricity from grid",
            # "Heating output energy of DHW heat pump",
            # "Relative electricity demand of DHW heat pump",
            "Residents' total electricity consumption",
            "Residents' electricity consumption from grid",
            # "Relative electricity demand of residents",
            # "Battery charging energy",
            # "Battery discharging energy",
            # "Battery losses",
        ]
    )
    ELECTRICITY_DATA = DataInfo(
        descriptions=[
            # "ElectricityMeter|Electricity|ElectricityToGrid",
            # "ElectricityMeter|Electricity|ElectricityFromGrid",
            "ElectricityMeter|Electricity|ElectricityToAndFromGrid",
            # "ElectricityMeter|Electricity|ElectricityConsumption",
            "UTSPConnector|Electricity|ElectricityOutput",
            "PVSystem_w0|Electricity|ElectricityOutput",
            "AdvancedHeatPumpHPLib|Electricity|ElectricalInputPower",
            "DHWHeatPump_w1|Electricity|ElectricityOutput",
            "ElectricCar_1_w1|Electricity|ElectricityOutput",
            "ElectricCar_2_w1|Electricity|ElectricityOutput",
            "ElectricCar_3_w1|Electricity|ElectricityOutput",
            "ElectricCar_4_w1|Electricity|ElectricityOutput",
            "ElectricCar_5_w1|Electricity|ElectricityOutput",
            "ElectricCar_6_w1|Electricity|ElectricityOutput",
            "ElectricCar_7_w1|Electricity|ElectricityOutput",
            "ElectricCar_8_w1|Electricity|ElectricityOutput",
            "ElectricCar_9_w1|Electricity|ElectricityOutput",
            "ElectricCar_10_w1|Electricity|ElectricityOutput",
            "ElectricCar_11_w1|Electricity|ElectricityOutput",
            "ElectricCar_12_w1|Electricity|ElectricityOutput",
        ]
    )
    FLOW_AND_RETURN_TEMPERTAURES = DataInfo(
        descriptions=[
            "AdvancedHeatPumpHPLib|Heating|TemperatureOutput",
            "SimpleHotWaterStorage|Water|WaterTemperatureToHeatGenerator",
            "SimpleHotWaterStorage|Water|WaterTemperatureToHeatDistribution",
            "HeatDistributionSystem|Water|WaterTemperatureOutput",
            "Weather|Temperature|DailyAverageOutsideTemperatures",
        ]
    )
    OCCUPANCY_CONSUMPTION = DataInfo(
        descriptions=["Occupancy|Electricity|ElectricityOutput", "Occupancy|WarmWater|WaterConsumption"]
    )
    HEATING_DEMAND = DataInfo(
        descriptions=["AdvancedHeatPumpHPLib|Heating|ThermalOutputPower", "Building|Temperature|TemperatureIndoorAir"]
    )
    VARIABLES_FOR_DEBUGGING_PURPOSES = DataInfo(
        descriptions=[
            "AdvancedHeatPumpHPLib|Heating|ThermalOutputPower",
            "AdvancedHeatPumpHPLib|Electricity|ElectricalInputPowerForHeating",
            "AdvancedHeatPumpHPLib|Electricity|ElectricalInputPowerForCooling",
            "Building|Temperature|TemperatureIndoorAir",
            "AdvancedHeatPumpHPLib|Any|COP",
            "Weather|Temperature|DailyAverageOutsideTemperatures",
            "HeatDistributionController|Temperature|HeatingFlowTemperature",
        ]
    )


class ScenarioAggregationEnumClass(Enum):

    """Scenario Aggregation Enum class is for deciding according to which scenarios the data will be aggregated."""

    BUILDING_TYPE = DataInfo(descriptions=["SFH", "TH", "MFH", "AB"])
    BUILDING_TYPE_AND_AGE = DataInfo(
        descriptions=[
            # "DE.N.TH.10",
            # "DE.N.TH.09",
            # "DE.N.TH.08",
            # "DE.N.TH.07",
            # "DE.N.TH.06",
            # "DE.N.TH.05",
            # "DE.N.TH.04",
            # "DE.N.TH.03",
            # "DE.N.TH.02",
            # "DE.N.SFH.10",
            # "DE.N.SFH.09",
            # "DE.N.SFH.08",
            # "DE.N.SFH.07",
            # "DE.N.SFH.06",
            # "DE.N.SFH.05",
            # "DE.N.SFH.04",
            # "DE.N.SFH.03",
            # "DE.N.SFH.02",
            # "DE.N.SFH.01",
            "DE.N.MFH.10",
            "DE.N.MFH.09",
            "DE.N.MFH.08",
            "DE.N.MFH.07",
            "DE.N.MFH.06",
            "DE.N.MFH.05",
            "DE.N.MFH.04",
            "DE.N.MFH.03",
            "DE.N.MFH.02",
            "DE.N.MFH.01",
            "DE.N.AB.06",
            "DE.N.AB.05",
            "DE.N.AB.04",
            "DE.N.AB.03",
            "DE.N.AB.02",
            "DE.East.MFH.05",
            "DE.East.MFH.04",
            "DE.East.AB.08",
            "DE.East.AB.07",
            "DE.East.AB.06",
        ]
    )
    BUILDING_REFURBISHMENT_STATE = DataInfo(descriptions=["001.001", "001.002", "001.003"])

    BUILDING_AGE = DataInfo(
        descriptions=[
            "01.Gen",
            "02.Gen",
            "03.Gen",
            "04.Gen",
            "05.Gen",
            "06.Gen",
            "07.Gen",
            "08.Gen",
            "09.Gen",
            "10.Gen",
            "11.Gen",
            "12.Gen",
        ]
    )

    def return_enum_values_as_dict(self) -> Dict[str, List[str]]:
        """Return enum values as dict."""
        dict_to_return: Dict[str, List[str]] = {"building_code": self.value.descriptions}
        return dict_to_return



"""Data Collection for Scenario Comparison."""
# clean
import time
import os
import sys
from typing import List, Optional, Dict
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hisim.postprocessing.scenario_evaluation import (
    result_data_collection,
    result_data_processing,
    result_data_plotting,
)
from hisim.component import ConfigBase
from hisim import log


@dataclass_json
@dataclass
class ScenarioAnalysisConfig(ConfigBase):

    """Configuration for running a scenario analysis."""

    building_name: str
    name: str
    data_format_type: str
    time_resolution_of_data_set: str
    cluster_storage_path: str
    module_results_directory: str
    result_folder_description_one: str
    result_folder_description_two: str
    path_to_default_module_config: str
    data_processing_mode: str
    simulation_duration_to_check: str
    variables_to_check: List[str]
    dict_with_scenarios_to_check: Optional[Dict]
    dict_with_extra_information_for_specific_plot: Dict[str, Dict]

    @classmethod
    def get_default(cls):
        """Get default ScenarioAnalysisConfig."""

        return ScenarioAnalysisConfig(
            building_name="BUI1",
            name="ScenarioAnalysisConfig_0",
            data_format_type=result_data_processing.DataFormatEnum.CSV.name,
            time_resolution_of_data_set=result_data_processing.ResultDataTypeEnum.YEARLY.name,
            cluster_storage_path="system_setups/",
            module_results_directory="results/household_cluster_advanced_hp_pv_battery_ems/",
            result_folder_description_one="PV-1-hds-2-hpc-mode-2/",
            result_folder_description_two="weather-location-BAD_MARIENBURG",
            path_to_default_module_config="/fast/home/k-rieck/jobs_hisim/cluster-hisim-paper/job_array_for_hisim_mass_simus/default_config_for_builda_data.json",
            data_processing_mode=result_data_collection.ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name,
            simulation_duration_to_check=str(365),
            variables_to_check=result_data_processing.OutputVariableEnumClass.KPI_DATA.value.descriptions,
            dict_with_scenarios_to_check=None,
            dict_with_extra_information_for_specific_plot={
                "scatter": {
                    "x_data_variable": "Conditioned floor area"
                },  # "Building|Temperature|TemperatureIndoorAir"     "Specific heating demand according to TABULA" "Weather|Temperature|DailyAverageOutsideTemperatures"
                "stacked_bar": {
                    "y1_data_variable": "Mean flow temperature of heat pump",
                    "y2_data_variable": "Mean return temperature of heat pump",
                    "use_y1_as_bottom_for_y2": False,
                    "sort_according_to_y1_or_y2_data": "y2",
                },
            },
        )


class ScenarioAnalysisWithConfig:

    """ScenarioAnalysis class which executes result data collection, processing and plotting."""

    def __init__(self, scenario_analysis_config: ScenarioAnalysisConfig) -> None:
        """Initialize the class."""
        # Get input parameters from config
        try:
            config_name = scenario_analysis_config.name.split("_")[1]
        except Exception:
            config_name = ""

        data_processing_mode = scenario_analysis_config.data_processing_mode
        data_format_type = scenario_analysis_config.data_format_type
        folder_from_which_data_will_be_collected = os.path.join(
            *[
                scenario_analysis_config.cluster_storage_path,
                scenario_analysis_config.module_results_directory,
                scenario_analysis_config.result_folder_description_one,
                scenario_analysis_config.result_folder_description_two,
            ]
        )
        path_to_default_config = scenario_analysis_config.path_to_default_module_config
        time_resolution_of_data_set = scenario_analysis_config.time_resolution_of_data_set
        simulation_duration_to_check = scenario_analysis_config.simulation_duration_to_check
        variables_to_check = scenario_analysis_config.variables_to_check
        dict_with_scenarios_to_check = scenario_analysis_config.dict_with_scenarios_to_check
        dict_with_extra_information_for_specific_plot = (
            scenario_analysis_config.dict_with_extra_information_for_specific_plot
        )

        result_data_collection_instance = result_data_collection.ResultDataCollection(
            data_processing_mode=data_processing_mode,
            scenario_analysis_config_name=config_name,
            data_format_type=data_format_type,
            folder_from_which_data_will_be_collected=folder_from_which_data_will_be_collected,
            path_to_default_config=path_to_default_config,
            time_resolution_of_data_set=time_resolution_of_data_set,
            simulation_duration_to_check=simulation_duration_to_check,
        )
        result_data_plotting.ScenarioChartGeneration(
            simulation_duration_to_check=simulation_duration_to_check,
            filepath_of_aggregated_dataframe=result_data_collection_instance.filepath_of_aggregated_dataframe,
            scenario_config_name=config_name,
            data_format_type=data_format_type,
            time_resolution_of_data_set=time_resolution_of_data_set,
            data_processing_mode=data_processing_mode,
            variables_to_check=variables_to_check,
            dict_of_scenarios_to_check=dict_with_scenarios_to_check,
            dict_with_extra_information_for_specific_plot=dict_with_extra_information_for_specific_plot,
        )


def main():
    """Main function to execute the scenario analysis."""

    # Get inputs for scenario analysis
    python_arguments = sys.argv

    if len(python_arguments) == 1:
        use_default_scenario_analysis_config: bool = True

    elif len(python_arguments) == 2:
        use_default_scenario_analysis_config = False
        scenario_analysis_config_path = sys.argv[1]
    else:
        raise ValueError(
            f"There should be 1 or 2 python arguments (sys.argv). Here {len(python_arguments)} are given. Please check your code."
        )

    my_config: ScenarioAnalysisConfig
    if use_default_scenario_analysis_config is False:
        if isinstance(scenario_analysis_config_path, str) and os.path.exists(
            scenario_analysis_config_path.rstrip("\r")
        ):
            with open(
                scenario_analysis_config_path.rstrip("\r"), encoding="unicode_escape"
            ) as scenario_analysis_config_file:

                my_config = ScenarioAnalysisConfig.from_json(scenario_analysis_config_file.read())  # type: ignore

            log.information(f"Read scenario analysis config from {scenario_analysis_config_path}")
            log.information("Config values: " + f"{my_config.to_dict}" + "\n")
        else:
            # cannot open file for scenario analysis config so default config will be used
            use_default_scenario_analysis_config = True

    if use_default_scenario_analysis_config is True:
        my_config = ScenarioAnalysisConfig.get_default()

        log.information("No scenario analysis config path was given or could be opened. Default config is used.")

    # -------------------------------------------------------------------------------------------------------------------------------------

    ScenarioAnalysisWithConfig(scenario_analysis_config=my_config)


if __name__ == "__main__":
    start_time = time.time()
    main()
    print(f"---{time.time() - start_time} seconds ___")



"""Csvloader module."""

# clean

import os
from typing import List
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import pandas as pd


from hisim import loadtypes as lt
from hisim import utils
from hisim import component as cp
from hisim.simulationparameters import SimulationParameters


@dataclass_json
@dataclass
class CSVLoaderConfig(cp.ConfigBase):
    """Csvloader config class."""

    building_name: str
    name: str
    csv_filename: str
    column: int
    loadtype: lt.LoadTypes
    unit: lt.Units
    column_name: str
    sep: str
    decimal: str
    multiplier: float
    output_description: str

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return CSVLoader.get_full_classname()


class CSVLoader(cp.Component):
    """Csvloader class.

    Class component loads CSV file containing some
    load profile relevant to the applied setup
    function.

    Parameters
    ----------
    name: str
        Name of load profile from CSV file
    csv_filename: str
        Name of CSV filename containing the load profile data
    column: int
        Column number where the load profile data is stored
        inside of the CSV file
    loadtype: LoadTypes,
        Load type corresponded to the data loaded
    unit: lt.Units
        Units of data loaded
    column_name: str
        Name of column where the load profile data is stored
        inside of the CSV File
    simulation_parameters: cp.SimulationParameters
        Simulation parameters used by the setup function
    sep: str
        Separator used CSV file
    decimal: str
        Decimal indicator used in the CSV file
    multiplier: float
        Multiplication factor, in case an amplification of
        the data is required

    """

    Output1: str = "CSV Profile"

    def __init__(
        self,
        config: CSVLoaderConfig,
        my_simulation_parameters: SimulationParameters,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ):
        """Initialize the class."""
        self.csvconfig = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.output1_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.Output1,
            self.csvconfig.loadtype,
            self.csvconfig.unit,
            output_description="CSV loader output 1",
        )
        self.output1_channel.display_name = self.csvconfig.column_name
        self.multiplier = self.csvconfig.multiplier

        # ? self.column = column
        dataframe = pd.read_csv(
            os.path.join(utils.HISIMPATH["inputs"], self.csvconfig.csv_filename),
            sep=self.csvconfig.sep,
            decimal=self.csvconfig.decimal,
        )
        if self.csvconfig.column >= len(dataframe.columns):
            raise RuntimeError(
                f"Invalid column number for the csv file: {self.csvconfig.column}. Found {len(dataframe.columns)} columns."
            )
        dfcolumn = dataframe.iloc[:, [self.csvconfig.column]]
        self.column_name = self.csvconfig.column_name
        if len(dfcolumn) < self.my_simulation_parameters.timesteps:
            raise Exception(
                "Timesteps: "
                + str(self.my_simulation_parameters.timesteps)
                + " vs. Lines in CSV "
                + self.csvconfig.csv_filename
                + ": "
                + str(len(self.column_name))
            )

        self.column = dfcolumn.to_numpy(dtype=float)
        self.values: List[float] = []

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        stsv.set_output_value(self.output1_channel, float(self.column[timestep]) * self.multiplier)

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        return self.csvconfig.get_string_dict()



"""The Simulator class forms the framework for all HiSim simulations.

It iterates over all components in each timestep until convergence and loops over all time steps.
"""
# clean
import os
import datetime
from typing import List, Tuple, Optional, Dict, Any, Union
import time
import pandas as pd

from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer
from hisim.component_wrapper import ComponentWrapper
from hisim import sim_repository
from hisim.postprocessing import postprocessing_main as pp
import hisim.component as cp
import hisim.dynamic_component as dcp
from hisim import log
from hisim.simulationparameters import SimulationParameters
from hisim import utils
from hisim import postprocessingoptions
from hisim.loadtypes import Units
from hisim.result_path_provider import ResultPathProviderSingleton, SortingOptionEnum


__authors__ = "Noah Pflugradt, Vitor Hugo Bellotto Zago, Maximillian Hillen"
__copyright__ = "Copyright 2020-2022, FZJ-IEK-3"
__license__ = "MIT"
__maintainer__ = "Noah Pflugradt"
__email__ = "n.pflugradt@fz-juelich.de"
__status__ = "production"


class Simulator:

    """Core class of HiSim: Runs the main loop."""

    @utils.measure_execution_time
    def __init__(
        self,
        module_directory: str,
        module_filename: str,
        my_simulation_parameters: Optional[SimulationParameters],
        setup_function: str = "setup_function",
        my_module_config: Optional[str] = None,
    ) -> None:
        """Initializes the simulator class and creates the result directory."""

        self._simulation_parameters: SimulationParameters
        if my_simulation_parameters is not None:
            self._simulation_parameters = my_simulation_parameters
            log.LOGGING_LEVEL = self._simulation_parameters.logging_level
        self.wrapped_components: List[ComponentWrapper] = []
        self.all_outputs: List[cp.ComponentOutput] = []

        self.setup_function = setup_function
        self.module_filename = module_filename
        self.module_directory = module_directory
        self.my_module_config = my_module_config
        self.simulation_repository = sim_repository.SimRepository()
        self.results_data_frame: pd.DataFrame
        self.iteration_logging_path: str = ""
        self.config_dictionary: Dict[str, Any] = {}

    def set_simulation_parameters(self, my_simulation_parameters: SimulationParameters) -> None:
        """Sets the simulation parameters and the logging level at the same time."""
        self._simulation_parameters = my_simulation_parameters
        if self._simulation_parameters is not None:
            log.LOGGING_LEVEL = self._simulation_parameters.logging_level

    def add_component(
        self,
        component: cp.Component,
        is_cachable: bool = False,
        connect_automatically: bool = False,
    ) -> None:
        """Adds component to simulator and wraps it up the output in the register."""
        if self._simulation_parameters is None:
            raise ValueError("Simulation Parameters were not initialized")
        # set the repository
        component.set_sim_repo(self.simulation_repository)

        # set the wrapper
        wrap = ComponentWrapper(component, is_cachable, connect_automatically=connect_automatically)
        wrap.register_component_outputs(self.all_outputs)
        self.wrapped_components.append(wrap)
        if component.component_name in self.config_dictionary:
            raise ValueError("duplicate component name : " + component.component_name)
        self.config_dictionary[component.component_name] = component.config

    @utils.measure_execution_time
    def connect_all_components(self) -> None:
        """Connects the inputs from every component to the corresponding outputs."""
        for wrapped_component in self.wrapped_components:
            wrapped_component.connect_inputs(self.all_outputs)

    @utils.measure_execution_time
    def prepare_calculation(self) -> None:
        """Connects the inputs from every component to the corresponding outputs."""
        for wrapped_component in self.wrapped_components:
            # check if component should be connected to default connections automatically
            if wrapped_component.connect_automatically is True:
                self.connect_everything_automatically(
                    source_component_list=[wp.my_component for wp in self.wrapped_components],
                    target_component=wrapped_component.my_component,
                )
            wrapped_component.prepare_calculation()

    def process_one_timestep(
        self, timestep: int, previous_stsv: cp.SingleTimeStepValues
    ) -> Tuple[cp.SingleTimeStepValues, int, bool]:
        """Executes one simulation timestep.

        Some components can be connected in a circle.
        To solve the circular dependency, all components have their states restored
        and simulated until their values converge.

        Firstly, their previously converged state is saved as the current timestep state.
        Following up, all components have their states restored and simulated respectively.
        Convergence is dependent on the i_restore and i_simulate of the components and how they
        are connected to each other.
        """

        # Save states of all components
        # Executes save state in the component
        for wrapped_component in self.wrapped_components:
            wrapped_component.save_state()

        continue_calculation = True

        # Verifies data existence
        if (len(self.all_outputs)) == 0:
            raise ValueError("Not a single column was defined.")

        # Creates List with values
        stsv = previous_stsv.clone()
        # Creates a buffer List with values
        previous_values = previous_stsv.clone()
        iterative_tries = 0
        force_convergence = False

        # Starts loop
        while continue_calculation:
            # Loops through components
            for wrapped_component in self.wrapped_components:
                # Executes restore state for each component
                wrapped_component.restore_state()
                # Executes i_simulate for component
                wrapped_component.calculate_component(timestep, stsv, force_convergence)

            # Stops simulation for too small difference between
            # actual values and previous values
            if stsv.is_close_enough_to_previous(previous_values):
                continue_calculation = False
            if (
                iterative_tries > 2
                and postprocessingoptions.PostProcessingOptions.PROVIDE_DETAILED_ITERATION_LOGGING
                in self._simulation_parameters.post_processing_options
            ):
                myerr = stsv.get_differences_for_error_msg(previous_values, self.all_outputs)
                with open(self.iteration_logging_path, "a", encoding="utf-8") as filestream:
                    filestream.write(myerr + "\n")
            if iterative_tries > 10:
                force_convergence = True
            if iterative_tries > 100:
                list_of_changed_values = stsv.get_differences_for_error_msg(previous_values, self.all_outputs)
                raise ValueError("More than 100 tries in time step " + str(timestep) + "\n" + list_of_changed_values)
            # Copies actual values to previous variable
            previous_values.copy_values_from_other(stsv)
            iterative_tries += 1

        for wrapped_component in self.wrapped_components:
            wrapped_component.doublecheck(timestep, stsv)
        return (stsv, iterative_tries, force_convergence)

    def prepare_simulation_directory(self):
        """Prepares the simulation directory. Determines the filename if nothing is set."""

        if (
            self._simulation_parameters.result_directory is None
            or len(self._simulation_parameters.result_directory) == 0
            or self._simulation_parameters.result_directory == ""
        ):

            # check if result path is already set somewhere manually
            if ResultPathProviderSingleton().get_result_directory_name() is not None:
                self._simulation_parameters.result_directory = ResultPathProviderSingleton().get_result_directory_name()
                log.information(
                    "Using result directory: "
                    + self._simulation_parameters.result_directory
                    + " which is set manually."
                )
            else:
                # if not, build a flat result path itself
                ResultPathProviderSingleton().set_important_result_path_information(
                    module_directory=self.module_directory,
                    model_name=self.module_filename,
                    variant_name=None,
                    scenario_hash_string=None,
                    sorting_option=SortingOptionEnum.FLAT,
                )
                self._simulation_parameters.result_directory = ResultPathProviderSingleton().get_result_directory_name()
                log.information(
                    f"Using result directory:  {self._simulation_parameters.result_directory}"
                    + " which is set by the simulator."
                )

        if not os.path.isdir(self._simulation_parameters.result_directory):
            os.makedirs(self._simulation_parameters.result_directory, exist_ok=True)

        log.LOGGING_LEVEL = self._simulation_parameters.logging_level
        self.iteration_logging_path = os.path.join(
            self._simulation_parameters.result_directory, "Detailed_Iteration_Log.txt"
        )

    # @profile
    # @utils.measure_execution_time
    def run_all_timesteps(self) -> None:
        """Performs all the timesteps of the simulation and saves the results in the attribute results."""
        # Error Tests
        # Test if all parameters were initialized
        if self._simulation_parameters is None:
            raise ValueError("Simulation Parameters were not initialized")

        # Tests if wrapper has any components at all
        if len(self.wrapped_components) == 0:
            raise ValueError("Not a single component was defined. Quitting.")

        # call again because it might not have gotten executed depending on how it's called.
        self.prepare_simulation_directory()

        flagfile = os.path.join(self._simulation_parameters.result_directory, "finished.flag")
        if self._simulation_parameters.skip_finished_results and os.path.exists(flagfile):
            log.warning("Found " + flagfile + ". This calculation seems finished. Quitting.")
            return
        # Starts time counter
        start_counter = time.perf_counter()
        self.prepare_calculation()
        # Connects all components
        self.connect_all_components()
        log.information(
            "finished connecting all components. A total of "
            + str(len(self.wrapped_components))
            + " components were defined. They have a total of "
            + str(len(self.all_outputs))
            + " outputs."
        )
        all_result_lines = []
        log.information("Starting simulation for " + str(self._simulation_parameters.timesteps) + " timesteps")
        lastmessage = datetime.datetime.now()
        last_step: int = 0
        starttime = datetime.datetime.now()
        total_iteration_tries_since_last_msg = 0

        # Creates empty list with values to get started
        number_of_outputs = len(self.all_outputs)
        stsv = cp.SingleTimeStepValues(number_of_outputs)

        for step in range(self._simulation_parameters.timesteps):
            if self._simulation_parameters.timesteps % 500 == 0:
                log.information("Starting step " + str(step))

            (
                resulting_stsv,
                iteration_tries,
                force_convergence,
            ) = self.process_one_timestep(step, stsv)
            stsv = cp.SingleTimeStepValues(number_of_outputs)
            # Accumulates iteration counter
            total_iteration_tries_since_last_msg += iteration_tries

            # Appends
            all_result_lines.append(resulting_stsv.values)
            del resulting_stsv
            # Calculates time execution
            elapsed = datetime.datetime.now() - lastmessage

            # For simulation longer than 5 seconds
            if elapsed.total_seconds() > 5 and step != 0:
                lastmessage = self.show_progress(
                    starttime,
                    step,
                    total_iteration_tries_since_last_msg,
                    last_step,
                    force_convergence,
                )
                last_step = step
                total_iteration_tries_since_last_msg = 0
        postprocessing_datatransfer = self.prepare_post_processing(all_result_lines, start_counter)
        log.information("Starting postprocessing")
        if postprocessing_datatransfer is None:
            raise ValueError("postprocessing_datatransfer was none")

        my_post_processor = pp.PostProcessor()
        my_post_processor.run(ppdt=postprocessing_datatransfer)
        for wrapped_component in self.wrapped_components:
            wrapped_component.clear()
        del all_result_lines
        del postprocessing_datatransfer
        del my_post_processor
        self.simulation_repository.clear()
        log.information("Finished postprocessing")
        with open(flagfile, "a", encoding="utf-8") as filestream:
            filestream.write("finished")

    @utils.measure_execution_time
    def prepare_post_processing(self, all_result_lines, start_counter):
        """Prepares the post processing."""
        log.information("Preparing post processing")
        # Prepares the results from the simulation for the post processing.
        if len(all_result_lines) != self._simulation_parameters.timesteps:
            raise ValueError("not all lines were generated")
        colum_names = []
        if self.setup_function is None:
            raise ValueError("No setup function was set")
        entry: cp.ComponentOutput
        for _index, entry in enumerate(self.all_outputs):
            column_name = entry.get_pretty_name()
            colum_names.append(column_name)
            log.debug("Output column: " + column_name)
        self.results_data_frame = pd.DataFrame(data=all_result_lines, columns=colum_names)
        # todo: fix this constant
        df_index = pd.date_range("2021-01-01 00:00:00", periods=len(self.results_data_frame), freq="T")
        self.results_data_frame.index = df_index
        end_counter = time.perf_counter()
        execution_time = end_counter - start_counter
        log.information(f"Simulation took {execution_time:1.2f}s.")
        (
            results_merged_cumulative,
            results_merged_monthly,
            results_merged_daily,
            results_merged_hourly,
        ) = self.get_std_results(self.results_data_frame)

        ppdt = PostProcessingDataTransfer(
            results=self.results_data_frame,
            all_outputs=self.all_outputs,
            simulation_parameters=self._simulation_parameters,
            wrapped_components=self.wrapped_components,
            mode=1,
            setup_function=self.setup_function,
            module_filename=self.module_filename,
            my_module_config=self.my_module_config,
            execution_time=execution_time,
            results_monthly=results_merged_monthly,
            results_cumulative=results_merged_cumulative,
            results_hourly=results_merged_hourly,
            results_daily=results_merged_daily,
        )
        log.information("Finished preparing post processing")
        return ppdt

    def show_progress(
        self,
        starttime: datetime.datetime,
        step: int,
        total_iteration_tries: int,
        last_step: int,
        force_covergence: bool,
    ) -> datetime.datetime:
        """Makes the pretty progress messages with time estimate."""
        # calculates elapsed time
        elapsed = datetime.datetime.now() - starttime
        elapsed_minutes, elapsed_seconds = divmod(elapsed.seconds, 60)
        elapsed_seconds_str: str = str(elapsed_seconds).zfill(2)
        # Calculates steps achieved per time duration
        steps_per_second = step / elapsed.total_seconds()
        elapsed_steps: int = step - last_step
        if elapsed_steps == 0:
            average_iteration_tries: float = 1
        else:
            average_iteration_tries = total_iteration_tries / elapsed_steps
        time_elapsed = datetime.timedelta(seconds=(self._simulation_parameters.timesteps - step) / steps_per_second)
        time_left_minutes, time_left_seconds = divmod(time_elapsed.seconds, 60)
        time_left_seconds = str(time_left_seconds).zfill(2)  # type: ignore
        simulation_status = f"Simulating... {(step / self._simulation_parameters.timesteps) * 100:.1f}% "
        simulation_status += f"| Elapsed Time: {elapsed_minutes}:{elapsed_seconds_str} min "
        simulation_status += f"| Speed: {steps_per_second:.0f} step/s "
        simulation_status += f"| Time Left: {time_left_minutes}:{time_left_seconds} min"
        simulation_status += f"| Avg. iterations {average_iteration_tries:.1f}"
        if force_covergence:
            simulation_status += " (forced)"
        log.information(simulation_status)
        return datetime.datetime.now()

    def get_std_results(
        self, results_data_frame: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Converts results into a pretty dataframe for post processing."""
        pd_timeline = pd.date_range(
            start=self._simulation_parameters.start_date,
            end=self._simulation_parameters.end_date,
            freq=f"{self._simulation_parameters.seconds_per_timestep}S",
        )[:-1]
        n_columns = results_data_frame.shape[1]

        results_data_frame.index = pd_timeline
        results_merged_monthly = pd.DataFrame()
        results_merged_daily = pd.DataFrame()
        results_merged_hourly = pd.DataFrame()
        results_merged_cumulative = pd.DataFrame()
        for i_column in range(n_columns):
            temp_df = pd.DataFrame(
                results_data_frame.values[:, i_column],
                index=pd_timeline,
                columns=[results_data_frame.columns[i_column]],
            )

            if self.all_outputs[i_column].unit in (
                Units.CELSIUS,
                Units.KELVIN,
                Units.ANY,
                Units.METER_PER_SECOND,
                Units.DEGREES,
                Units.WATT,
                Units.KILOWATT,
                Units.WATT_PER_SQUARE_METER,
                Units.KG_PER_SEC,
                Units.PERCENT,
                Units.PASCAL
            ):
                temp_df_monthly = temp_df.resample("M").mean()
                temp_df_daily = temp_df.resample("D").mean()
                temp_df_cumulative = temp_df.mean()
            else:
                temp_df_monthly = temp_df.resample("M").sum()
                temp_df_daily = temp_df.resample("D").sum()
                temp_df_cumulative = temp_df.sum()

            # monthly results
            results_merged_monthly[temp_df_monthly.columns[0]] = temp_df_monthly.values[:, 0]
            results_merged_monthly.index = temp_df_monthly.index
            # daily results
            results_merged_daily[temp_df_daily.columns[0]] = temp_df_daily.values[:, 0]
            results_merged_daily.index = temp_df_daily.index

            # cumulative results
            results_merged_cumulative[temp_df_monthly.columns[0]] = temp_df_cumulative.values

            if self._simulation_parameters.seconds_per_timestep != 3600:
                if self.all_outputs[i_column].unit in (
                    Units.CELSIUS,
                    Units.KELVIN,
                    Units.ANY,
                    Units.METER_PER_SECOND,
                    Units.DEGREES,
                    Units.WATT,
                    Units.KILOWATT,
                    Units.WATT_PER_SQUARE_METER,
                    Units.KG_PER_SEC,
                    Units.PERCENT,
                ):
                    temp_df_hourly = temp_df.resample("60T").mean()  # .interpolate(method="linear")
                else:
                    temp_df_hourly = temp_df.resample("60T").sum()

                results_merged_hourly[temp_df_hourly.columns[0]] = temp_df_hourly.values[:, 0]
                results_merged_hourly.index = temp_df_hourly.index
            else:
                results_merged_hourly[temp_df.columns[0]] = temp_df.values[:, 0]

        return (
            results_merged_cumulative,
            results_merged_monthly,
            results_merged_daily,
            results_merged_hourly,
        )

    def connect_everything_automatically(
        self,
        source_component_list: Union[List[cp.Component], List[dcp.DynamicComponent]],
        target_component: Union[cp.Component, dcp.DynamicComponent],
    ) -> None:
        """Connect chosen target component in the sytem setups automatically based on its default connections."""

        # prepare the target components' default connection lists
        target_default_connection_dict: Union[
            Dict[str, List[cp.ComponentConnection]],
            Dict[str, List[dcp.DynamicComponentConnection]],
        ]

        # check if target component is a normal or a dynamic component and get all default connections
        if isinstance(target_component, dcp.DynamicComponent):
            target_default_connection_dict = target_component.dynamic_default_connections

        elif isinstance(target_component, cp.Component) and not isinstance(target_component, dcp.DynamicComponent):
            target_default_connection_dict = target_component.default_connections

        else:
            raise TypeError(
                f"Type {type(target_component)} of target_component should be Component or Dynamic Component."
            )

        # check if target component has any default connections (otherwise automatic connection cannot be made)
        if bool(target_default_connection_dict) is True:
            # check if at least one source_component is in the target default connections
            if (
                any(
                    source_component.get_classname() in target_default_connection_dict
                    for source_component in source_component_list
                )
                is False
            ):
                raise KeyError(
                    f"No component in the system setup matches the default connections of {target_component.component_name}."
                )

            # go through all registered components
            for source_component in source_component_list:
                source_component_classname = source_component.get_classname()

                # if the source components' classname is found in the target components' default connection dict, a connection is made
                if source_component_classname in target_default_connection_dict.keys():
                    if isinstance(target_component, dcp.DynamicComponent):
                        dynamic_connections = target_component.get_dynamic_default_connections(
                            source_component=source_component
                        )

                        target_component.connect_with_dynamic_connections_list(
                            dynamic_component_connections=dynamic_connections
                        )

                    if isinstance(target_component, cp.Component) and not isinstance(
                        target_component, dcp.DynamicComponent
                    ):
                        connections = target_component.get_default_connections(source_component=source_component)
                        target_component.connect_with_connections_list(connections=connections)
        else:
            raise KeyError(
                f"Automatic connection does not work for {target_component.component_name} because no default connections were found. "
                + "Please check if a connection is needed and if yes, create the missing default connection in your component."
            )

    def put_log_files_into_result_path(self) -> None:
        """Put logging files from /logs path into result path."""

        default_logging_path = log.LOGGING_DEFAULT_PATH
        result_directory = self._simulation_parameters.result_directory

        if os.path.exists(result_directory) is False:
            raise NameError(f"The result directory {result_directory} could not be found.")

        # move log files to result path
        for file in os.listdir(default_logging_path):
            # if logging file is not yet in result directory, move it from default directory /logs to result directory
            if not os.path.isfile(os.path.join(result_directory, file)):
                os.rename(os.path.join(default_logging_path, file), os.path.join(result_directory, file))
